/*
	File:		FilterModule.c

	Contains:	Sample TPI STREAMS Module to filter data

		Not complete with respect to opening/closing, etc. Does not set TPI
		states correctly at all times.
		
*/

#ifndef __OPENTPTMODULE__
#include <OpenTptModule.h>
#endif
#ifndef _MPS_STROPTS_
#include <stropts.h>
#endif
#ifndef _TIHDR_
#include <tihdr.h>
#endif

/*
 * Forward function declarations
 */
 
int		filter_admin(void);
int		filter_close(queue_t*, int, cred_t*);
int		filter_open(queue_t*, dev_t*, int, int, cred_t*);
int		filter_rsrv(queue_t*);
int		filter_wsrv(queue_t*);
int		filter_rput(queue_t*, mblk_t*);
int		filter_wput(queue_t*, mblk_t*);

int		filter_wproto(queue_t* wrq, mblk_t*);
int		filter_wdata(queue_t* wrq, mblk_t*);
int		filter_rproto(queue_t* rdq, mblk_t*);
int		filter_rdata(queue_t* rdq, mblk_t*);
int		filter_error(queue_t* rdq, mblk_t*, int error);
void 	ReplyErrorAck(queue_t* rdq, mblk_t*, long reqType,
					  int xtiErr, int macErr);
/*
 * Module information for the streamtab structure
 */
 
static struct module_info minfo =  
{
	9001,				/* Module Number			*/
	"SmplFilter",		/* Name of module			*/
	0,					/* Minimum data size		*/
	INFPSZ,				/* Maximum data size		*/
	2048,				/* Hi water mark for queue	*/
	128					/* Lo water mark for queue	*/
};

/*
 * Info for the "read" queue.
 */
 
static struct qinit rinit = 
{
	filter_rput,		/* Put routine for "incoming" data			*/
	filter_rsrv,		/* Service routine for "incoming" data		*/
	filter_open,		/* Our open routine							*/
	filter_close, 		/* Our close routine						*/
	filter_admin,		/* Our admin routine						*/
	&minfo				/* Our module_info							*/
};

/*
 * Info for the "write" queue
 */
 
static struct qinit winit =
{
	filter_wput,		/* Put routine for client data					*/
	filter_wsrv,		/* Service routine for client data				*/
	0,					/* open routine goes in read-side structure 	*/
	0,					/* close routine goes in read-side structure	*/
	0,					/* admin routine goes in read-side structure	*/
	&minfo				/* Our module_info								*/
};

/*
 * The streamtab structure
 */
 
static struct streamtab theStreamTab = 
{
	&rinit,			/* Our read-side qinit structure	*/
	&winit,			/* Our write-side qinit structure	*/
	0,				/* Not a mux - no structure info	*/
	0				/* Not a mux - no structure info	*/
};

/*******************************************************************************
** Open routine
********************************************************************************/

	typedef struct error_struct
	{
		long		reqType;
		queue_t*	rdq;
		int			xtiErr;
		int			macErr;
		Boolean		allocd;
	} error_struct;
	
	typedef struct filter_struct
	{
		int				myError;
		UInt16			myState;
		error_struct	myErrorStruct;
		Boolean			errInUse;
	} filter_struct;
	
int filter_open(queue_t* rdq, dev_t* dev, int flag, int sflag, cred_t* creds)
{
#pragma unused (flag)
	int				err;
	filter_struct*	myData;
	/*
	 * Make sure caller is properly credentialed
	 */
	if ( (err = drv_priv(creds)) != kOTNoError )
		return err;
	/*
	 * If we're being reopened - just return
	 */
	if ( rdq->q_ptr != NULL )
		return kOTNoError;
	/*
	 * Make sure we're being opened properly - We only allow
	 * a "module" open.
	 */
	if ( sflag != MODOPEN )
		return ENXIO;
		
	myData = (filter_struct*)OTAllocMem(sizeof(filter_struct));
	if ( myData == NULL )
		return ENOSR;
		
	/*
	 * Set my state in my data structure
	 */
	myData->myState					= T_UNBND;
	myData->errInUse				= false;
	myData->myErrorStruct.allocd	= false;
	/*
	 * Store our data structure into the q_ptr field of the queue.
	 */
	rdq->q_ptr		= (char*)myData;
	WR(rdq)->q_ptr	= (char*)myData;

	/*
	 * return no error
	 */
	return kOTNoError;
}

/*******************************************************************************
** Close routine
********************************************************************************/

int filter_close(queue_t* rdq, int flags, cred_t* credP)
{
	/*
	 * Free the memory we stored at the q_ptr.
	 * Zero it out
	 */
	filter_struct* myData = (filter_struct*)rdq->q_ptr;
	OTFreeMem(rdq->q_ptr);
	rdq->q_ptr		= NULL;	
	WR(rdq)->q_ptr	= NULL;
	/*
	 * Return no error
	 */
	return kOTNoError;
}

/*******************************************************************************
** Admin routine
**
** Admin routines are for future expansion - just return no error.
********************************************************************************/

int filter_admin()
{
	return kOTNoError;
}

/*******************************************************************************
** Write-side put routine
********************************************************************************/

int filter_wput(queue_t* q, mblk_t* mp)
{
	/*
	 * Process the incoming message from the client
	 */
	switch (mp->b_datap->db_type)
	{
		/*
		 * These messages can't come from upstream
		 */
		case M_COPYIN:
		case M_COPYOUT:
		case M_ERROR:
		case M_HANGUP:
		case M_IOCACK:
		case M_IOCNAK:
		case M_SETOPTS:
		case M_SIG:
			freemsg(mp);
			return 0;
		/*
		 * Process a signal message - Under Open Transport, this is normally
		 * a timer message.  If not a timer message for myself, pass it on.
		 */
		case M_PCSIG:
			break;
		/*
		 * Undocumented message type - just pass it on.
		 */
		case M_HPDATA:
			break;
		/*
		 * You will get one of these if you use M_COPYIN.  If it's not for
		 * yourself, pass it on.
		 */
		case M_IOCDATA:
			break;
		/*
		 * These are messages for future expansion.  If not understood,
		 * pass them on
		 */
		case M_PCRSE:
		case M_RSE:
			break;
			
		/*
		 * Request to flush queues.  Do the job, then pass the message on
		 */
		case M_FLUSH:
			if ( mp->b_rptr[0] & FLUSHW)
			{
				if ( (mp->b_rptr[0] & FLUSHBAND) && mp->b_wptr - mp->b_rptr == 2 )
					flushband(q, mp->b_rptr[1], FLUSHALL);
				else
					flushq(q, FLUSHALL);
			}
			if ( mp->b_rptr[0] & FLUSHR )
			{
				if ( (mp->b_rptr[0] & FLUSHBAND) && mp->b_wptr - mp->b_rptr == 2 )
					flushband(RD(q), mp->b_rptr[1], FLUSHALL);
				else
					flushq(RD(q), FLUSHALL);
			}
			break;
			
		/*
		 * These messages are rarely uses, but you may need to process them.
		 */
		case M_STOP:
		case M_START:
		case M_STOPI:
		case M_STARTI:
		case M_READ:
		case M_BREAK:
		case M_DELAY:
		case M_CTL:
			break;
			
		/*
		 * Here, we process any IOCTLs that we might support.  If the IOCTL
		 * is not for uss, pass it on.
		 */
		case M_IOCTL:
			break;
		
		/*
		 * Handle data coming from the client
		 */
		case M_DATA:
			return filter_wdata(q, mp);
		
		/*
		 * Handle M_PROTO and M_PCPROTO requests coming from the client
		 */
		case M_PROTO:
		case M_PCPROTO:
			return filter_wproto(q, mp);
			
		/*
		 * Should never see one of these messages. It is an invisible message
		 * used by the streamhead.
		 */
		case M_PASSFP:
			break;
			
		default:
			break;
	}
	putnext(q, mp);
	return 0;
}

/*******************************************************************************
** Read-side put routine
********************************************************************************/

int filter_rput(queue_t* q, mblk_t* mp)
{
	/*
	 * Process the incoming message from the driver below us
	 */
	switch (mp->b_datap->db_type)
	{
		/*
		 * These shouldn't come from downstream
		 */
		case M_READ:
		case M_IOCDATA:
			freemsg(mp);
			return 0;
			
		/*
		 * Miscellaneous messages - pass them on if you don't want them
		 */
		case M_STOP:
		case M_START:
		case M_STOPI:
		case M_STARTI:
			break;
			
		/*
		 * Just send these on up to the streamhead
		 */
		case M_COPYIN:
		case M_COPYOUT:
		case M_ERROR:
		case M_HANGUP:
		case M_PCRSE:
			break;
		
		/*
		 * If we send any IOCTLs downstream, look at the acks and 
		 * nacks here to determine if these are ours.  If so, process
		 * them.  Otherwise, just send them on upstream
		 */
		case M_IOCACK:
		case M_IOCNAK:
			break;
		
		/*
		 * If we use any timers, process them here.  Otherwise, send
		 * the signal upstream
		 */
		case M_PCSIG:
			break;
		
		/*
		 * Undocumented message type - send it on
		 */
		case M_HPDATA:
			break;
			
		/*
		 * Request to flush queues.  Do the job, then pass the message on
		 */
		case M_FLUSH:
			if ( mp->b_rptr[0] & FLUSHW)
			{
				if ( (mp->b_rptr[0] & FLUSHBAND) && mp->b_wptr - mp->b_rptr == 2 )
					flushband(WR(q), mp->b_rptr[1], FLUSHALL);
				else
					flushq(WR(q), FLUSHALL);
			}
			if ( mp->b_rptr[0] & FLUSHR )
			{
				if ( (mp->b_rptr[0] & FLUSHBAND) && mp->b_wptr - mp->b_rptr == 2 )
					flushband(q, mp->b_rptr[1], FLUSHALL);
				else
					flushq(q, FLUSHALL);
			}
			break;
			
		/*
		 * Handle data coming from downstream
		 */
		case M_DATA:
			return filter_rdata(q, mp);
		
		/*
		 * Handle M_PROTO and M_PCPROTO requests coming from downstream
		 */
		case M_PROTO:
		case M_PCPROTO:
			return filter_rproto(q, mp);
			
		/*
		 * Anything we don't understand - just pass it on
		 */
		default:
			break;
	}
	putnext(q, mp);
	return 0;
}

/*******************************************************************************
** ReuseMessage
**
** Determine if we can reuse an mblk_t*.  If not, reallocate one.
********************************************************************************/

mblk_t* ReuseMessage(mblk_t* mp, size_t size)
{
	if ( mp != NULL )
	{
		if ( mp->b_cont != NULL )
		{
			freemsg(mp->b_cont);
			mp->b_cont = NULL;
		}
		mp->b_rptr = mp->b_datap->db_base;
		mp->b_wptr = mp->b_rptr;
		if ( mp->b_datap->db_ref == 1 &&
			 mp->b_datap->db_lim - mp->b_rptr >= size )
			return mp;
		freemsg(mp);
	}
	return allocb(size, BPRI_LO);
}

/*******************************************************************************
** filter_error
**
** Process a fatal error
********************************************************************************/

	static void retry_error(long arg)
	{
		filter_struct* myData = (filter_struct*)((queue_t*)arg)->q_ptr;
		(void)filter_error((queue_t*)arg, NULL, myData->myError);
	}
	
int filter_error(queue_t* rdq, mblk_t* mp, int error)
{
	filter_struct* myData = (filter_struct*)rdq->q_ptr;
	myData->myState = TS_BAD_STATE;
	/*
	 * Try to reuse the incoming message block.
	 */
	mp = ReuseMessage(mp, 1);
	/*
	 * If we can't get one, use bufcall to try this again
	 * when memory is available.
	 */
	if ( mp == NULL )
	{
		myData->myError = error;
		bufcall(1, BPRI_HI, (bufcall_t)retry_error, (long)rdq);
		return 0;
	}
	mp->b_wptr = mp->b_rptr + 1;
	mp->b_rptr[0] = (unsigned char)error;
	mp->b_datap->db_type = M_ERROR;
	putnext(rdq, mp);
}

/*******************************************************************************
** ReplyErrorAck
********************************************************************************/

	static void retry_ack(long arg)
	{
		error_struct* es = (error_struct*)arg;
		int	xtiErr	= es->xtiErr;
		int macErr	= es->macErr;
		long reqType	= es->reqType;
		queue_t* rdq	= es->rdq;
		if ( es->allocd )
			OTFreeMem(es);
		ReplyErrorAck(rdq, NULL, reqType, xtiErr, macErr);
	}
	
void ReplyErrorAck(queue_t* rdq, mblk_t* mp, long reqType, int xtiErr, int macErr)
{
	filter_struct* myData = (filter_struct*)rdq->q_ptr;
	mp = ReuseMessage(mp, sizeof(struct T_error_ack));
	if ( mp == NULL )
	{
		/*
		 * If we can't get a message block, either allocate an error_struct
		 * or use the one we have stored in our data.
		 */
		error_struct* es;
		if ( myData->errInUse )
		{
			es = (error_struct*)OTAllocMem(sizeof(error_struct*));
			if ( es != NULL )
				es->allocd = true;
		}
		else
		{
			myData->errInUse	= true;
			es = &myData->myErrorStruct;
		}
		/*
		 * If we could get an error_struct, try using bufcall.  If that fails -
		 * we're hosed.  Use filter_error to kill the stream
		 */
		if ( es != NULL )
		{
			es->reqType	= reqType;
			es->xtiErr	= xtiErr;
			es->macErr	= macErr;
			es->rdq		= rdq;
			if ( bufcall(sizeof(struct T_error_ack), BPRI_HI, (bufcall_t)retry_ack,
						 (long)es) != 0 )
			{
				if ( es->allocd )
					OTFreeMem(es);
				es = NULL;
			}
		}
		if ( es == NULL )
			(void)filter_error(rdq, NULL, ENOSR);
	}
	mp->b_wptr = mp->b_rptr + sizeof(struct T_error_ack);
	((struct T_error_ack*)mp->b_rptr)->ERROR_prim = reqType;
	((struct T_error_ack*)mp->b_rptr)->PRIM_type = T_ERROR_ACK;
	((struct T_error_ack*)mp->b_rptr)->TLI_error = xtiErr;
	((struct T_error_ack*)mp->b_rptr)->UNIX_error = macErr;
	mp->b_datap->db_type = M_PCPROTO;
	putnext(rdq, mp);
}

/*******************************************************************************
** Write-side processing routines
********************************************************************************/

int filter_wdata(queue_t* q, mblk_t* mp)
{
	filter_struct* myData = (filter_struct*)q->q_ptr;
	if ( myData->myState != TS_DATA_XFER && myData->myState != TS_WREQ_ORDREL )
	{
		/*
		 * TPI spec says if we're in the idle state, just ignore the message.
		 * Otherwise, it is a fatal error.
		 */
		if ( myData->myState == TS_IDLE )
		{
			freemsg(mp);
			return 0;
		}
		return filter_error(RD(q), mp, EPROTO);
	}
	
	/*	---------------------------------------------------------------------
		Here, we do whatever munging we're going to do on the client data,
		then send it on.
		Keep in mind that we call this routine from the filter_wproto routine,
		so that the first block of the message might be an M_PROTO, and not
		data.
		--------------------------------------------------------------------- */

	 
	 /*
	  * Now, we send it on.  If we can't because of flow control, use
	  * putq to put it on our queue and send it on.
	  */
	 if ( !bcanput(q->q_next, mp->b_band) )
	 	putq(q, mp);
	else
		putnext(q, mp);
	return 0;
}

int filter_wproto(queue_t* q, mblk_t* mp)
{
	Boolean			doOutState = false;
	filter_struct*	myData = (filter_struct*)q->q_ptr;
	/*
	 * If the message block is not big enough - blow it away
	 */
	if ( mp->b_wptr - mp->b_rptr < sizeof(long) )
	{
		freemsg(mp);
		return 0;
	}
	/*
	 * Look at the incoming primitive message
	 */
	switch ( ((union T_primitives*)mp->b_rptr)->type )
	{
		/*
		 * For incoming data, let filter_wdata handle it
		 */
		case T_DATA_REQ:
		case T_EXDATA_REQ:
			return filter_wdata(q, mp);
			
		case T_BIND_REQ:
			/*
			 * If we're not in the unbound state, give a
			 * TOUTSTATE error.  Otherwise, change our
			 * state to TS_WACK_BREQ to flag that we're in
			 * the process of binding, and pass the bind on.
			 */
			if ( myData->myState != TS_UNBND )
				doOutState = true;
			else
				myData->myState = TS_WACK_BREQ;
			break;
			
		case T_UNBIND_REQ:
			/*
			 * If we're not in the TS_UNBND or TS_WACK_BREQ states,
			 * give a TOUTSTATE error.  Otherwise, pass the unbind on.
			 */
			if ( myData->myState != TS_UNBND && myData->myState != TS_WACK_BREQ )
				doOutState = true;
			break;
			
		case T_OPTMGMT_REQ:
			/*
			 * We don't do option management, so just pass it on
			 */
			break;
			
		case T_CONN_REQ:
		case T_CONN_RES:
		case T_DISCON_REQ:

		/*
		 * These we just pass on to the layer below us
		 */
		case T_INFO_REQ:
		case T_ADDR_REQ:
		case T_RESOLVEADDR_REQ:
			break;
			
		/*
		 * We don't handle these - give an error if we get one
		 */
		case T_UNITDATA_REQ:
		case T_UREQUEST_REQ:
		case T_UREPLY_REQ:
		case T_CANCELREQUEST_REQ:
			return filter_error(RD(q), mp, EPROTO);
	}
	if ( doOutState )
	{
		ReplyErrorAck(RD(q), mp, ((union T_primitives*)mp->b_rptr)->type,
					  TOUTSTATE, 0);
		return 0;
	}
	/*
	 * Here, we just pass the message block on to the next layer, or
	 * queue it up till flow control is lifted.
	 */
	if (mp->b_datap->db_type > QPCTL || canput(q->q_next) )
		putnext(q, mp);
	else
		putq(q, mp);
	return 0;
}

int filter_wsrv(queue_t* q)
{
	mblk_t*	mp;
	/*
	 * Send on downstream anything that was scheduled for flow control!
	 */
	while ( (mp = getq(q)) != NULL )
	{
		/*
		 * If we're flow controlled, get out and we'll come back when the flow-
		 * control lifts.
		 */
		if ( !bcanput(q->q_next, mp->b_band) )
		{
			putbq(q, mp);
			return 0;
		}
		/*
		 * Otherwise, send the data downstream, and get some more
		 */
		putnext(q, mp);
	}
}

/*******************************************************************************
** Read-side process routines
********************************************************************************/

int filter_rdata(queue_t* q, mblk_t* mp)
{
	filter_struct* myData = (filter_struct*)q->q_ptr;
	/*
	 * If we're not ready for incoming data, just free it up
	 */
	if ( myData->myState != TS_DATA_XFER && myData->myState != TS_WREQ_ORDREL )
	{
		freemsg(mp);
		return 0;
	}
	
	/*	---------------------------------------------------------------------
		Here, we do whatever munging we're going to do on the incoming data,
		then send it on.
		Keep in mind that we call this routine from the filter_rproto routine,
		so that the first block of the message might be an M_PROTO, and not
		data.
		--------------------------------------------------------------------- */

	 
	 /*
	  * Now, we send it on.  If we can't because of flow control, use
	  * putq to put it on our queue and send it on later.
	  */
	 if ( !bcanput(q->q_next, mp->b_band) )
	 	putq(q, mp);
	else
		putnext(q, mp);
	return 0;
}

int filter_rproto(queue_t* q, mblk_t* mp)
{
	Boolean			doOutState = false;
	filter_struct*	myData = (filter_struct*)q->q_ptr;
	/*
	 * If the message block is not big enough - blow it away
	 */
	if ( mp->b_wptr - mp->b_rptr < sizeof(long) )
	{
		freemsg(mp);
		return 0;
	}
	/*
	 * Look at the incoming primitive message
	 */
	switch ( ((union T_primitives*)mp->b_rptr)->type )
	{
		/*
		 * These are just passed on, since we don't deal with them
		 */
		case T_INFO_ACK:
		case T_ADDR_ACK:
		case T_RESOLVEADDR_ACK:
		case T_OPTMGMT_ACK:
			break;
		
		/*
		 * These, we should never get.  If we do - we are totally confused
		 * and probably on a STREAM we shouldn't be on
		 */
		case T_UREPLY_ACK:
		case T_UNITDATA_IND:
		case T_UDERROR_IND:
		case T_UREQUEST_IND:
		case T_UREPLY_IND:
			return filter_error(q, mp, EPROTO);
			
		/*
		 * Here, a bind request succeeded.
		 */
		case T_BIND_ACK:
			myData->myState = TS_IDLE;
			break;
			
		/*
		 * Handle any error acks that come back
		 */
		case T_ERROR_ACK:
		{
			switch ( ((struct T_error_ack*)mp->b_rptr)->ERROR_prim )
			{
				/*
				 * These are error acks for functions we just pass thru
				 */
				case T_INFO_REQ:
				case T_ADDR_REQ:
				case T_OPTMGMT_REQ:
					break;
				
				/*
				 * If the unbind failed, then we are not unbound either.
				 * Send the error on up to the client.
				 */
				case T_UNBIND_REQ:
					myData->myState = TS_IDLE;
					break;
					
				/*
				 * If the bind failed, we fall back to the unbound state
				 */
				case T_BIND_REQ:
					myData->myState = TS_UNBND;
					break;
					
				case T_CONN_REQ:
				case T_CONN_RES:
				case T_DISCON_REQ:
				/*
				 * %%% Needs implementation
				 */
				 break;
				 
				/*
				 * Any error acks we don't understand just pass on
				 */
				default:
					break;
			}
			break;
		}
		
		/*
		 * Handle any OK acks that come back
		 */
		case T_OK_ACK:
		{
			switch ( ((struct T_ok_ack*)mp->b_rptr)->CORRECT_prim )
			{
				/*
				 * If the unbind is OK, we are no longer unbound.
				 */
				case T_UNBIND_REQ:
					myData->myState = TS_UNBND;
					break;
										
				case T_CONN_REQ:
				case T_CONN_RES:
				case T_DISCON_REQ:
				/*
				 * %%% Needs implementation - Kill any that have their
				 *     own acking mechanism.
				 */
				 break;
				 
				/*
				 * Any error acks we don't understand just pass on
				 */
				default:
					break;
			}
			break;
		}
		
		case T_CONN_IND:
		case T_CONN_CON:
		case T_DISCON_IND:
		case T_ORDREL_IND:
		/*
		 * %%% Needs implementation
		 */
		 break;
		
		/*
		 * Process data in the filter_rdata routine
		 */
		case T_DATA_IND:
		case T_EXDATA_IND:
			return filter_rdata(q, mp);
			
		/*
		 * Anything else we don't understand, just send it on
		 */
		default:
			break;
	}	/*
	 * Here, we just pass the message block on to the next layer, or
	 * queue it up till flow control is lifted.
	 */
	if (mp->b_datap->db_type > QPCTL || canput(q->q_next) )
		putnext(q, mp);
	else
		putq(q, mp);
	return 0;
}

int filter_rsrv(queue_t* q)
{
	mblk_t*	mp;
	/*
	 * Send on up anything that was scheduled for flow control!
	 */
	while ( (mp = getq(q)) != NULL )
	{
		/*
		 * If we're flow controlled, get out and we'll come back when the flow-
		 * control lifts.
		 */
		if ( !bcanput(q->q_next, mp->b_band) )
		{
			putbq(q, mp);
			return 0;
		}
		/*
		 * Otherwise, send the data upstream, and get some more
		 */
		putnext(q, mp);
	}
}

/* ------------------------------------------------------------------------

	THESE FUNCTIONS AND DATA STRUCTURES ARE SPECIFIC TO THE MACINTOSH IMPLEMENTATION
	OF STREAMS. THESE ROUTINES WOULD NOT EXIST IN OTHER STREAMS ENVIRONMENTS.
	
	WHEN PORTING OTHER STREAMS CODE TO THE MACINTOSH, THESE ROUTINES AND DATA
	STRUCTURES MUST BE ADDED.

  ------------------------------------------------------------------------ */

/*******************************************************************************
** install_info structure, and the GetInstallInfo function
**
** This is the structure that tells Open Transport about your module. OT calls
** your "GetInstallInfo" function to get this information.
**
********************************************************************************/

static struct install_info theInstallInfo =
{
	&theStreamTab,			// Stream Tab pointer
	kOTModIsModule, 		// Tell OT that we are a module, not a driver
	SQLVL_MODULE,			// Synchronization level - the module will never
							// be re-entered with this setting
	0,						// Shared writer list buddy
	0,						// Open Transport use - always set to 0
	0						// Flag - always set to 0
};
	
install_info* GetOTInstallInfo()
{
	return &theInstallInfo;
}

/*******************************************************************************
** This is the function that will be called the first time that this code
** is loaded.  It is guaranteed that this function is call at SystemTask time
** on the Macintosh.
********************************************************************************/

Boolean InitStreamModule(void* cookie)
{
	/* Do any machine specific initialization stuff you need to */;
	return true;
}

